@@ -2,6 +2,7 @@ from __future__ import print_function
 
 from datetime import datetime
 from warnings import warn
+from signal import SIGUSR1, signal
 from socket import AF_INET, SOCK_STREAM, socket
 import argparse
 import os
@@ -11,7 +12,7 @@ import subprocess
 import sys
 
 from . import settings
-from .diskemptysampler import DiskEmptySampler
+from .disksampler import DiskSampler
 from .hdddock import BaseStorageDevice
 from .inventoryrestclient import InventoryRESTClient
 from .labelprinterapiclient import get_labelprinter
@@ -50,6 +51,10 @@ class Inventory:
         return (self._data.get('current_health') or {}).get('status')
 
 
+class ActionInterrupt(Exception):
+    pass
+
+
 class SymlinkNotFound(Exception):
     pass
 
@@ -161,13 +166,13 @@ class DeviceManager(object):
             'DEBUG: Sampling block device to check if empty... '
             '(sample_count: {}, sample_size {})'.format(
                 sample_count, sample_size))
-        sampler = DiskEmptySampler(
+        sampler = DiskSampler(
             self._hdd_dock.devname, sample_count, sample_size)
-        sampler.sample()
-        if sampler.is_zero():
-            print('DEBUG: No non-zero sample found. Disk already zeroed?')
-        else:
-            print('DEBUG: Found non-zero samples before wipe (expected)')
+        # Write a fixed sample and read it back from disk.
+        sampler.write_sample()
+        if sampler.is_different():
+            print('WARNING: written sample does not match read sample! '
+                  '(possibly faulty disk)')
 
         print('Start secure wiping BLKDEV')
         success, error = erase_method()
@@ -215,15 +220,6 @@ class DeviceManager(object):
         finally:
             self._set_health('SECURE ERASED')
 
-    def _find_non_zero_blocks(self, sample_count, sample_size):
-
-        disk_empty_sampler = DiskEmptySampler(
-            self._hdd_dock.devname, sample_count, sample_size)
-        result = disk_empty_sampler.sample_disk()
-        assert result is not None
-
-        return result is False  # false = we found non-zero
-
     def led_failure(self):
         self._hdd_dock.led_failure()
 
@@ -330,6 +326,15 @@ def registered_disk_actions(auto_action=None, owner='', **kwargs):
         hdd_dock, hdd_id, inventory_rest_client, author, location)
     manager.show_summary(_build_print_data())
 
+    # Setup signal handler for remove actions.
+    remove_actions = [i[1:] for i in auto_action if i.startswith('/')]
+    if remove_actions:
+        auto_action = [i for i in auto_action if not i.startswith('/')]
+
+        def remove_disk_interrupt(signum, frame):
+            raise ActionInterrupt(remove_actions)
+        signal(SIGUSR1, remove_disk_interrupt)
+
     is_auto = bool(auto_action)
     show_commands = bool(not is_auto)
 
@@ -345,7 +350,13 @@ def registered_disk_actions(auto_action=None, owner='', **kwargs):
             if is_auto:
                 manager.show_commands()
                 is_auto = False
-            action = input('\nAction: ').upper()
+            try:
+                action = input('\nAction: ').upper()
+            except ActionInterrupt as interrupt:
+                print('\n\n\x1b[1mDISK REMOVED\x1b[0m')
+                auto_action = interrupt.args[0]
+                is_auto = True
+                continue
 
         if action in ('1', 'set-owner'):
             while len(owner) == 0:
@@ -697,7 +708,7 @@ class AutoServer:
         # It's possible a device does not eject/spin down properly and triggers
         # the udev rules again. Keep a list of disk serials which have been
         # started and ignore subsequent notifications about that disk.
-        self.seen_devices = set()
+        self.seen_devices = {}
         if not gui:
             if not os.environ.get('TMUX'):
                 args = [
@@ -750,11 +761,7 @@ class AutoServer:
             print('Waiting for disk notifications...')
             while True:
                 connection, address = sock.accept()
-                pid = os.fork()
-                if pid == 0:
-                    self.handle_client(connection)
-                else:
-                    os.waitid(os.P_PID, pid, os.WEXITED)
+                self.handle_client(connection)
 
     def handle_client(self, connection):
         client_data = b''
@@ -770,15 +777,25 @@ class AutoServer:
             try:
                 cmd, device = notification.split(':', 1)
             except ValueError:
-                print(f'Received bogus notification: {notification!r}')
+                print(f'Received malformed notification: {notification!r}')
                 return
 
             device = Path(device)
             if device.is_block_device():
-                if self.is_ignored(device, force=bool(cmd == 'device-force')):
-                    print(f'Ignoring notification for {device}')
-                else:
-                    self.start_docktool(device)
+                hdd_dock = BaseStorageDevice.from_devname(device)
+                serial = hdd_dock.get_serial_number()
+                if self.is_ignored(
+                        device, serial, force=bool(cmd.endswith('-force'))):
+                    print(f'Ignoring notification for {device}:{serial}')
+                elif cmd.startswith('device-add'):
+                    self.start_docktool(device, serial)
+            elif cmd.startswith('device-remove'):
+                # Block device no longer exists but if we have seen it before
+                # we will send the remove signal to the last osso-docktool
+                # for that device.
+                if device in self.seen_devices:
+                    self.seen_devices.pop(device)
+                    self.signal_docktool(device)
             else:
                 print(f'Path {device} is not a block device')
 
@@ -786,44 +803,48 @@ class AutoServer:
     def docktool_exec(self):
         return Path(sys.argv[0]).resolve()
 
-    def is_ignored(self, device, force):
-        hdd_dock = BaseStorageDevice.from_devname(device)
-        serial = hdd_dock.get_serial_number()
-        if not force and serial in self.seen_devices:
-            print(f'Device {device} with serial {serial} has been processed')
+    def is_ignored(self, device, serial, force):
+        if device.resolve() in self.ignore_devices:
+            return True
+
+        if not force and serial in self.seen_devices.values():
+            print(f'Skipped repeat Device {device} with serial {serial}')
             print(
                 f'Use "osso-docktool auto-notify {device} --force" to start')
             return True
-        self.seen_devices.add(serial)
-        # Check the resolved path against the ignore list.
-        return bool(device.resolve() in self.ignore_devices)
+        self.seen_devices[device] = serial
+        return False
 
-    def start_docktool(self, device):
+    def start_docktool(self, device, serial):
         if self.gui:
-            self.start_gui(device)
+            self.start_gui(device, serial)
         else:
-            self.start_cli(device)
+            self.start_cli(device, serial)
 
-    def start_gui(self, device):
-        print(f'Starting GUI terminal for {device}')
-        # Double fork to detach terminal from pid group.
+    def start_gui(self, device, serial):
+        print(f'Starting GUI terminal for {device}:{serial}')
+        pid = os.fork()
+        if pid > 0:
+            os.waitid(os.P_PID, pid, os.WEXITED)
+            return
+
+        # Double fork to detach X terminal from pid group.
         if os.fork() == 0:
-            args = [f'osso-docktool-terminal [{device}]', '-e']
+            args = [f'osso-docktool-terminal [{device}:{serial}]', '-e']
             args += self.get_docktool_cmd(device)
             os.setsid()
             os.execvp('x-terminal-emulator', args)
         else:
             os._exit(0)
 
-    def start_cli(self, device):
-        print(f'Creating new tmux window for {device}')
+    def start_cli(self, device, serial):
+        print(f'Creating new tmux window for {device}:{serial}')
         if not os.environ.get('TMUX'):
             print('Error: TMUX environment variable is empty', file=sys.stderr)
         else:
             args = ['tmux', 'new-window', '-n', device]
             args += self.get_docktool_cmd(device)
             subprocess.check_call(args)
-        os._exit(0)
 
     def get_docktool_cmd(self, device):
         cmd = ['sudo', '-E', self.docktool_exec, 'dev', device, '--wait']
@@ -833,10 +854,23 @@ class AutoServer:
             cmd += ['--owner', self.owner]
         return cmd
 
+    def signal_docktool(self, device):
+        # Perform on remove actions with the last active osso-docktool for the
+        # given device.
+        pid = subprocess.check_output(
+            ['pgrep', '-nf', f'osso-docktool dev {device}'], text=True).strip()
+        if pid.isdigit():
+            os.kill(int(pid), SIGUSR1)
+
+
+def auto_notify(device, notification, force):
+    if force:
+        notification = f'{notification}-force:{device}'
+    else:
+        notification = f'{notification}:{device}'
 
-def auto_notify(device, force=False):
-    notification = f'device-force:{device}' if force else f'device:{device}'
     with socket(AF_INET, SOCK_STREAM) as sock:
+        sock.settimeout(3)
         try:
             sock.connect(('127.0.0.1', 1451))
             sock.sendall(notification.encode())
@@ -851,6 +885,13 @@ class DocktoolArgumentParser(argparse.ArgumentParser):
             description='Docktool for processing disks')  # exit_on_error=False
         self.message2 = None
 
+        actions = [
+            'dispose', 'eject', 'print-label', 'print-health-label',
+            'locate', 'set-owner', 'quick-erase', 'secure-erase',
+            # Prefix with / to use as remove action.
+            '/print-health-label',
+        ]
+
         self.subparsers = self.add_subparsers(
             title='subcommands', dest='command', required=True,
             help='subcommand help',
@@ -863,10 +904,7 @@ class DocktoolArgumentParser(argparse.ArgumentParser):
             'device', metavar='DISK', type=self.block_device,
             help='Device to use for example: /dev/sda')
         devparser.add_argument(
-            '-a', '--action', choices=[
-                'dispose', 'eject', 'print-label', 'print-health-label',
-                'locate', 'set-owner', 'quick-erase', 'secure-erase'],
-            action='append', default=[],
+            '-a', '--action', choices=actions, action='append', default=[],
             help='Automatically execute the actions in order.')
         devparser.add_argument(
             '-o', '--owner', metavar='OWNER', default='',
@@ -900,21 +938,28 @@ class DocktoolArgumentParser(argparse.ArgumentParser):
         autoparser.add_argument(
             # https://bugs.python.org/issue27227
             # Add [] to choices to make actions optional.
-            'action', nargs='*', choices=[
-                'dispose', 'eject', 'print-label', 'print-health-label',
-                'locate', 'set-owner', 'quick-erase', 'secure-erase', []],
+            'action', nargs='*', choices=actions + [[]],
             help='Start docktool and perform these actions in order '
                  '(default: start osso-docktool, no extra action).')
 
         notifyparser = self.subparsers.add_parser(
             'auto-notify',
             help='Notify the auto server of the device')
+        notifyparser.add_argument(
+            '-a', '--add', action='store_const', const='device-add',
+            dest='notification', help='Send a device add notification.')
+        notifyparser.add_argument(
+            '-r', '--remove', action='store_const', const='device-remove',
+            dest='notification', help='Send a device remove notification.')
         notifyparser.add_argument(
             '-f', '--force', action='store_true',
             help='Force the auto server to process the device.')
+        # Notify can be used on removed disks which cannot validate
+        # as a block device type.
         notifyparser.add_argument(
-            'device', metavar='DISK', type=self.block_device,
+            'device', metavar='DISK',
             help='Device to use for example: /dev/sda')
+        notifyparser.set_defaults(notification='device-add')
 
     def parse_args(self):
         if len(sys.argv) == 2 and not (
@@ -971,7 +1016,7 @@ def main():
         elif args.command == 'auto':
             AutoServer(args.action, args.gui, args.ignore, args.owner).start()
         elif args.command == 'auto-notify':
-            auto_notify(args.device, args.force)
+            auto_notify(args.device, args.notification, args.force)
     except KeyboardInterrupt:
         sys.exit(0)
     except SystemExit:
