@@ -151,10 +151,11 @@ class SmartDataParser(object):
         # Default to 512 (bytes)
         return 512
 
-    def _get_value(self, search_str, column_name):
+    def _get_value(
+            self, search_str, column_name, search_column='attribute_name'):
         data = [
             x for x in self.attributes
-            if x.get('attribute_name', '').lower() == search_str]
+            if x.get(search_column, '').lower() == search_str]
 
         if data:
             value = data[0].get(column_name, None)
@@ -173,9 +174,14 @@ class SmartDataParser(object):
         # This is a percentage (the value, that is), which goes down from 100%
         # to 0% (= no health); taken from the obscurely named
         # "wear_leveling_count".
-        keys = ('media_wearout_indicator', 'wear_leveling_count')
-        for key in keys:
-            wear = self._get_value(key, 'value')
+        keys = (
+            ('media_wearout_indicator', 'attribute_name'),
+            ('wear_leveling_count', 'attribute_name'),
+            # Micron lifetime remaining is an unknown attr in smartctl 7.1
+            ('202', 'id#'),
+        )
+        for key, column in keys:
+            wear = self._get_value(key, 'value', column)
             if wear is not None:
                 break
         else:
